/**
 * @file Imagenes.c
 * @author Gregory Randall
 * @date 4 dec 2024
 * @brief Código fuente bilbioteca tratamiento de imagenes.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include "iio.h"
#include "Imagenes.h"

#define frand() ((double) rand() / RAND_MAX) /* da numero aleatorio entre o y 1 */
//#define DEBUG
#ifdef DEBUG
    #define DEBPRINT(...) fprintf(stderr, __VA_ARGS__)
#else
    #define DEBPRINT(...)
#endif

//#define DEBUG1
//#define DEBUG2
//#define DEBUG3
//#define DEBUG4

/**
 * @brief Agrega a imagen pin ruido sal y pimienta. Potencia indica el
 * porcentaje de ruido respecto al número de pixeles en la imagen
 * salida en pout.
 *
 * @param pin Puntero a imagen a la que agregar ruido.
 * @param pout Puntero a imagen con ruido.
 * @param potencia (flotante entre 0.0 y 1.0).
 * @param seed (int) semilla para el random generator
 * @return CodigoError.
 */
CodigoError AgregarRudioSalYPimienta(const Imagen* pin, Imagen* pout, float potencia, int seed){
  int i,k,l,size;
  float rn;
  srand(seed);
 
  duplicarImagen(pin,pout);
  
#ifdef DEBUG4
  printf("sal y pimienta\n");
  int w = pout->columnas;
  int h = pout->filas;
  printf("w %d, h %d, tipo %d\n",w,h,pout->tipo);
  iio_write_image_uint8_vec("toto.png", pout->pixeles, w, h, 1);
#endif
  size = (pin->filas * pin->columnas)*potencia;
  for (i=0; i<size; i++){
    rn = frand();
    k = (int) (rn*pin->filas);
    rn = frand();
    l = (int) (rn*pin->columnas);
    rn = frand();
    pout->pixeles[k*pin->columnas+l] = (uint8_t)(rn>0.5? 255 : 0);
  }
#ifdef DEBUG4
  iio_write_image_uint8_vec("totoRSP.png", pout->pixeles, w, h, 1);
#endif
  return OK; 
}

/**
 * @brief Generador de muestras Gaussianas. Método de Box-Muller Transform.
 * de https://phoxis.org/2013/05/04/generating-random-numbers-from-normal-distribution-in-c/
 * Se debe correr srand(seed) antes de llamar a esta funcion.
 * @param mu y sigma, parametros de la distribucion
 * @return Sample.
 */

double SampleGaussian(double mu, double sigma){
  double u,v,r,m;
  static double X1, X2;
  static int call =0;
  
  if(call ==1){
    call = !call;
    return(mu+sigma*(double)X2);
  }
  /* u y v deben estar en un rango entre 0 y 1 */
  do{
    u = ((double)rand()/RAND_MAX)*2 - 1.0; /* da numero aleatorio entre -1 y 1 */
    v = ((double)rand()/RAND_MAX)*2 - 1.0; /* da numero aleatorio entre -1 y 1 */
    r = u*u + v*v;
#ifdef DEBUG
    printf("u: %f v: %f r: %f\n",u,v,r);
#endif
  }while((r >= 1.0) || ( r == 0.0));
  m = sqrt(-2*log(r)/r);
  X1 = u*m;
  X2 = v*m;
  call =! call;
  return(mu+sigma*(double)X1);
}

/**
 * @brief Agrega a imagen pin ruido Gaussiano. Potencia indica el
 * porcentaje de ruido en términos de Std respecto al rango de 256 niveles de gris en imagen
 * salida en pout. Se incluye como parámetro la semilla para el generador de números aleatorios.
 
 * @param pin Puntero a imagen a la que agregar ruido.
 * @param pout Puntero a imagen con ruido.
 * @param potencia (flotante entre 0.0 y 1.0).
 * @param seed (int) semilla para el random generator
 * @return CodigoError.
 */

CodigoError AgregarRudioGaussiano(const Imagen* pin, Imagen* pout, float potencia, int seed){
  int i,j,ind;
  int val,nval;
  double Sample,Std,Mean;
  float ss,Sum = 0;
  float media;

  Mean = 0.0;
  Std = potencia * 256.0; /* si 256 es el maximo y potencia es un % de ese valor maximo */ 
  int n = pin->filas*pin->columnas;
  int * tmp = (int*) malloc(sizeof(int)*n);
  int menos,mas;
  menos = mas = 0;
  srand(seed);

#ifdef DEBUG
  printf("potencia %f Std %f\n",potencia,Std);
#endif
  
  duplicarImagen(pin,pout);
  for(i = 0; i < pin->filas; i++)
    for(j = 0; j < pin->columnas; j++){
      ind=i*pin->columnas+j;
      Sample=SampleGaussian(Mean, Std);
      val = (int)(Sample);
      tmp[ind]=val;
      Sum += val;
#ifdef DEBUG  
      printf("Sample %f val %d\n",Sample,val);
#endif
      nval = val + pout->pixeles[ind];
      if (nval < 0){
	pout->pixeles[ind] = 0;
	menos++;
      }
      else if (nval > 255){
	pout->pixeles[ind] = 255;
	mas++;
      }
      else
	pout->pixeles[ind] = (uint8_t)nval;
    }
  
  media = Sum/n;
  ss = 0;
  for(i = 0; i < pin->filas; i++)
    for(j = 0; j < pin->columnas; j++){
      val = tmp[i*pin->columnas+j];
      ss += (val - media)*(val - media);
    }
  free(tmp);
  printf("fil %d col %d n %d media %f ss %f Std %f Nb vals < 0 %d Nb vals > 0 %d\n",pin->filas,pin->columnas,n,media, ss, sqrt(ss/(n)),menos,mas);
  return OK; 
  
}


int  cmpfunc(const void *a, const void *b){
  unsigned char A = *(unsigned char*)a;
  unsigned char B = *(unsigned char*)b;
  return ((int)(A-B));
}

/**
 * @brief Filtra por mediana de radio orden salida en pout.
 *
 * @param pin Puntero a imagen a la que agregar ruido.
 * @param pout Puntero a imagen con ruido.
 * @param orden entero para calcular la ventana del filtro. ( 2*orden+1)^2 (entero).
 * @return CodigoError.
 */
CodigoError FiltroMediana(const Imagen* pin, Imagen* pout, int orden){
  int i,j,k,l,n,ind,kk;
  int r=2*orden+1;
  duplicarImagen(pin,pout);
  
/*ponemos a cero orden fila y orden columnas al princiipio y al final */
  printf("Mediana\n");
  
  unsigned char * aux = (unsigned char *)malloc(r*r*sizeof(unsigned char*));
#ifdef DEBUG
  printf("r %d pin->columnas %d pin-> filas %d\n",r,pin->columnas,pin->filas);
#endif
  for(i = 0; i < (pin->filas); i++){
    for(j = 0; j < (pin->columnas); j++){
      ind=i*pin->columnas+j;
      n=0;
#ifdef DEBUG
      printf("i %d j %d gris %d\n",i,j,pin->pixeles[ind]);
#endif
      for(k=i-orden;k<=i+orden;k++){
	for(l=j-orden;l<=j+orden;l++){
	  kk=k*pin->columnas+l;
	  aux[n++] = (unsigned char)pin->pixeles[kk];
	}
      }
      qsort(aux, r*r, sizeof(unsigned char),cmpfunc);
      if((i < orden) || (i >= pin->filas-orden) || (j < orden) || (j >= pin->columnas - orden))
          pout->pixeles[ind] = 0;
      else
          pout->pixeles[ind] = (uint8_t)aux[(r*r)/2];
#ifdef DEBUG
      for(l=0;l<9;l++)
	printf("%d ",aux[l]);
      printf("\n");
      printf("i %d j %d gris %d\n\n",i,j,pout->pixeles[ind]);
#endif
    }
  }
  
  free(aux);
  return OK; 
}

CodigoError FiltroPromedio(const Imagen* pin, Imagen* pout, int orden){
  int i,j,k,l,ind,kk;
  float val;
  int den=(2*orden+1)*(2*orden+1);
  duplicarImagen(pin,pout);

  printf("promedio\n");
  for(i = 0; i < (pin->filas); i++){
    for(j = 0; j < (pin->columnas); j++){
      ind=i*pin->columnas+j;
#ifdef DEBUG
      printf("i %d j %d gris %d\n",i,j,pin->pixeles[ind]);
#endif
      val=0;
      for(k=i-orden;k<=i+orden;k++){
	for(l=j-orden;l<=j+orden;l++){
	  kk=k*pin->columnas+l;
	  val += pin->pixeles[kk];
	}
      }
      if((i < orden) || (i >= pin->filas-orden) || (j < orden) || (j >= pin->columnas - orden))
         pout->pixeles[ind] = 0;
      else
         pout->pixeles[ind] = (uint8_t)(val/den);
#ifdef DEBUG
      printf("P> i %d j %d gris %d\n\n",i,j,pout->pixeles[ind]);
#endif
    }
  }
  
  return OK; 
}


/**
 * @brief Genera el histograma de pin y lo coloca en histo
 *
 * @param pin Puntero a imagen sobre la cual calcular el histograma
 * @param histo puntero a puntero de enetros, creada en la funcion
 * @return el histogrma.
 */
long int * Histo(const Imagen* pin){
  int i,j;

  long int *H =(long int*)calloc(sizeof(long int),256);
  
  for(i = 0; i < pin->filas; i++){
    for(j = 0; j < pin->columnas; j++){
      H[pin->pixeles[i*pin->columnas+j]]++;
    }
  }
  printf("Histograma\n");
  return H;
}



/**
 * @brief Genera el histograma accumulado a partir del histograma
 *
 * @param histoIn Puntero a histograma
 * @param histoAcum puntero a histograma acumulado, allocado internamente
 * @param Si vale distinto de 0, sus valores entre 0 y 255, para ser utilizado con LUT. En ese caso consideramos que el numero total de pixels de la imagen esta en el parametro.
 * @return histograma acumulado.
 */
long int * HistoAcum(const long int* histo, int size){
  int i;
  long int * Acum = (long int *) malloc(sizeof(long int)*256);
  
  Acum[0] = histo[0];
  for(i = 1; i < 256; i++)
    Acum[i] = histo[i] + Acum[i-1];
  
  if (size != 0){
    for(i = 0; i < 256; i++)
       Acum[i] = (Acum[i]*255)/size;
  }
  printf("Histograma acumulado\n");
  return Acum;
}

	   
void Plot_Histo(long int* histo, int nbBins, float factor){
  int i,j,step;
  long int max=0;
  int stepBin = 256/nbBins;
  
  for(i=0;i<255;i++)
    if (max < histo[i])
      max = histo[i];
  step = (factor == 0 ? 1 : (int)(max/(max*factor)));
  printf("Printf Histograma max %ld step %d stepBin %d\n",max,step,stepBin);

  long int tmp = 0;
  for(i=0;i<255;i++){
    tmp += histo[i];
    if( i % stepBin == 0){
      printf("%d \t",i);
      for(j=0;j<tmp;j++){
	if(j%step == 0)
	  printf("*");
      }
      printf("\n");
    }
    tmp = 0;
  }
}


/**
 * @brief Genera imagen a traves de una LUT
 *
 * @param pin Puntero a imagen sobre la cual operar
 * @param pout Puntero a imagen de salida
 * @param LUT puntero a Look Up Table
 * @return CodigoError.
 */
CodigoError LUT(const Imagen* pin, Imagen* pout, long int * LUT){
  int i,j,ind;

  if(LUT == NULL)
    return DATOS_INVALIDOS;

  duplicarImagen(pin,pout);
  
  for(i = 0; i < pin->filas; i++){
    for(j = 0; j < pin->columnas; j++){
      ind = i*pin->columnas+j;
      if(LUT[(int)pin->pixeles[ind]] > 255)
	 printf("error lut %d %d %d %ld\n",i,j,pin->pixeles[ind],LUT[(int)pin->pixeles[ind]]);
      pout->pixeles[ind] = (uint8_t) LUT[(int)pin->pixeles[ind]] ;
    }
  }

  return OK;
}

CodigoError imprimirHisto(long int *histo){
  if(histo == NULL)
    return DATOS_INVALIDOS;
  int i;
  printf("histo:\n");
  for(i=0;i<256;i++)
    printf("%ld ",histo[i]);
  printf("\n");
  return OK;
}

CodigoError imprimirRegionFloat(float *img,int fil, int col, int fi,int ff,int ci, int cf){
  if(img == NULL)
    return DATOS_INVALIDOS;
  int i,j,k=fi;
  printf("imprimir region:\n     ");
  for(i=ci;i<=cf;i++)
    printf("\t%3d",i);
  printf("\n");
  for(i=fi;i<=ff;i++){
    printf("%d : ",k++);
    for(j=ci;j<=cf;j++)
      printf("\t%0.1f",img[i*col+j]);
    printf("\n");
  }

  return OK;
}

CodigoError imprimirRegionInt(int *img,int fil, int col, int fi,int ff,int ci, int cf){
  
  if(img == NULL)
    return DATOS_INVALIDOS;
  int i,j,k=fi;
  printf("imprimir region:\n     ");
  for(i=fi;i<=ff;i++)
    printf("%3d ",i);
  printf("\n");
  for(i=fi;i<=ff;i++){
    printf("%d : ",k++);
    for(j=ci;j<=cf;j++)
      printf("%3d ",img[i*col+j]);
    printf("\n");
  }

  return OK;
}

CodigoError imprimirRegionImg(const Imagen* pin, int fi,int ff,int ci, int cf, int col){
  if(pin == NULL)
    return DATOS_INVALIDOS;
  int i,j,ind,k=fi;
  printf("imprimir region:\n");
  for(i=ci;i<=cf;i++)
    printf("\t%d",i);
  printf("\n");
  for(i=fi;i<=ff;i++){
    printf("%d : ",k++);
    for(j=ci;j<=cf;j++){
      ind = i*col+j;
      printf("\t%d", pin->pixeles[ind]);
    }
    printf("\n");
  }

  return OK;
}



/**
 * @brief Inicializa una imagen y reserva memoria para sus pixeles.
 *
 * Al tratarse "pixeles" de un puntero de punteros requiere reservar memoria 
 * en dos niveles: primero de tamaño filas*sizeof(puntero), obteniendo "filas" 
 * punteros a Pixel, y luego cada puntero a Pixel debe apuntar a "columnas" 
 * pixeles. Es decir, se tienen filas punteros de punteros, cada uno apuntando
 * a columnas "Pixel"es-
 *
 * @param pnew Puntero a imagen a inicializar
 * @param columnas Número de columnas de la imagen.
 * @param filas Número de filas de la imagen.
 * @param tipo Tipo imagen (color o gris).
 * @return CodigoError.
 */
CodigoError inicializarImagen(Imagen* pnew, int columnas, int filas, TipoImagen tipo){
   if(pnew!= NULL){
     pnew->tipo = tipo;
     pnew->columnas = columnas;
     pnew->filas = filas;
     pnew->pixeles = NULL;
     pnew->pixelesR = NULL;
     pnew->pixelesG = NULL;
     pnew->pixelesB = NULL;
#ifdef DEBUG4
     printf("en inicializar imagen w %d h %d tipo %d\n",pnew->columnas,pnew->filas,pnew->tipo);
#endif
     DEBPRINT("inicializarImagen: se retornó sin errores\n");
     return OK;
   }
   else{
     DEBPRINT("inicializarImagen: ERROR en parámetros de entrada.\n");
     return ERROR;
   }
}

 /**
 * @brief Libera memoria de pixeles y coloca atributos imagen a 0.
 *
 * Desarma el proceso explicado en inicializarImagen.
 *
 * @param pimg Puntero a imagen a liberar.
 * @return CodigoError.
 */
 CodigoError destruirImagen(Imagen* pimg){
   if(pimg!= NULL){
     pimg->tipo = 0;
     pimg->columnas = 0;
     pimg->filas = 0;
     if(pimg->pixeles != NULL)
       free(pimg->pixeles);
     if(pimg->tipo == COLOR){
       if(pimg->pixelesR != NULL)
	 free(pimg->pixelesR);
       if(pimg->pixelesG != NULL)
	 free(pimg->pixelesG);
       if(pimg->pixelesB != NULL)
	 free(pimg->pixelesB);
     }
     DEBPRINT("destruirImagen: se retornó sin errores\n");
     return OK;
   }
   else{
     DEBPRINT("destruirImagen: ERROR en parámetros de entrada.\n");
     return ERROR;
   }
 }

 /**
 * @brief Copia la inicializacion de la imagen apuntada por pin en pout.
 *
 * @param pin Puntero a imagen origen copia.
 * @param pout Puntero a imagen destino copia.
 * @return CodigoError.
 */
CodigoError duplicarImagen(const Imagen* pin, Imagen* pout){
  DEBPRINT("duplicarImagen: se inicializará la nueva imagen.\n");
  int w = pin->columnas;
  int h = pin->filas;
  int tipo = pin->tipo;
#ifdef DEBUG4
  printf("antes inicializar imagen w %d h %d tipo %d\n",w,h,tipo);
#endif
  inicializarImagen(pout, w, h, tipo);
#ifdef DEBUG4
  printf("en duplicar imagen w %d h %d tipo %d\n",pout->columnas,pout->filas,pout->tipo);
#endif
  if((pout->pixeles = (uint8_t *)malloc(w*h*sizeof(uint8_t))) == NULL)
    return ERROR_MEMORIA;
  if((pout->pixelesR = (uint8_t *)malloc(w*h*sizeof(uint8_t))) == NULL)
    return ERROR_MEMORIA;
  if((pout->pixelesG = (uint8_t *)malloc(w*h*sizeof(uint8_t))) == NULL)
    return ERROR_MEMORIA;
  if((pout->pixelesB = (uint8_t *)malloc(w*h*sizeof(uint8_t))) == NULL)
    return ERROR_MEMORIA;

  int i;
  for(i=0;i<w*h;i++){
    pout->pixeles[i] = pin->pixeles[i];
    if(pin->tipo == COLOR){
      pout->pixelesR[i] = pin->pixelesR[i];
      pout->pixelesG[i] = pin->pixelesG[i];
      pout->pixelesB[i] = pin->pixelesB[i];
    }
  }

  return OK;
}


/**
 * @brief Lee el contenido del archivo ubicado en ruta en la imagen apuntada
 * por pimg
 *
 * Es capaz de leer archivos en formato PPM, PPM plano, PGM, PGM plano.
 *
 * @param ruta arreglo de caracteres con la ruta al archivo que se desea leer.
 * @param pin Puntero a imagen para cargar la imagen.
 * @return CodigoError.
 */
CodigoError leerImagen(const char* ruta, Imagen* pimg){
  int w,h,pd;
  int i,j,k,ind,indi;

  uint8_t *ImgRGBi = iio_read_image_uint8_vec(ruta, &w, &h, &pd);
  pimg->tipo = (pd == 3 ? COLOR : GRISES);
  
  pimg->pixeles = (uint8_t*)malloc(w*h*sizeof(uint8_t));
  if(pimg->tipo == COLOR){
    pimg->pixelesR = (uint8_t*)malloc(w*h*sizeof(uint8_t));
    pimg->pixelesG = (uint8_t*)malloc(w*h*sizeof(uint8_t));
    pimg->pixelesB = (uint8_t*)malloc(w*h*sizeof(uint8_t));
  }
  //  printf("w %d h %d pd %d\n",w,h,pd);
  if(pd==1){
    for(i=0;i<w*h;i++)
      pimg->pixeles[i] = (uint8_t)ImgRGBi[i];
  }
  else{
    for(i = 0; i < h; i++){
      for(j = 0,k=0; j < w; j++,k+=pd){
	ind=i*w*pd+k;
	indi=i*w+j;
	pimg->pixelesR[indi] = ImgRGBi[ind];
	pimg->pixelesG[indi] = ImgRGBi[ind+1];
	pimg->pixelesB[indi] = ImgRGBi[ind+2];
      }
    }
  }
  //  printf("i %d j %d k %d\n",i,j,k);

  pimg->columnas = w;
  pimg->filas = h;
  free(ImgRGBi);
#ifdef DEBUG
  printf("en leer imagen w %d h %d pd %d\n",w,h,pd);
#endif
  
  return OK;
}

/**
 * @brief Guarda el contenido de la imagen apuntada por pimg en el archivo
 * ubicado en ruta.
 *
 * Es capaz de escribir archivos en formato PPM, PPM plano, PGM, PGM plano.
 *
 * @param pin Puntero a imagen que se desea escribir en archivo.
 * @param ruta Arreglo de caracteres con la ruta al archivo que se desea
 * escribir.
 *
 * @return CodigoError.
 */
CodigoError escribirImagen(const Imagen* pimg, char* ruta){
  if(pimg == NULL)
    return ERROR;
  int h = pimg->filas;
  int w = pimg->columnas;
  int pd;
  pd = (pimg->tipo == GRISES ? 1 : 3);

  if(pimg->pixeles == NULL)
    return ERROR;
  
  iio_write_image_uint8_vec(ruta, pimg->pixeles, w, h, pd);
  return OK;
}

/**
 * @brief copia los atributos de una imagen de tipo color de entrada apuntada
 * por pin en la imagen de tipo gris apuntada por pout.
 *
 * @param pin Puntero a imagen color
 * @param pout Puntero a imagen blanco y negro, puede no estar inicializado.
 * @return CodigoError.
 */
CodigoError convertirImagenColorAGris(const Imagen* pin, Imagen* pout){
  int i,j,errorduplicar;
  if(pin == NULL||pout == NULL){
    DEBPRINT("convertirImagenColorAGris: ERROR en parámetros.\n");
    return ERROR;
  }
  errorduplicar = duplicarImagen(pin,pout);
  if(errorduplicar){
    DEBPRINT("convertirImagenColorAGris: ERROR\n");
    return ERROR;
  }
  if(pin->tipo == GRISES){
    DEBPRINT("ya era gris\n");
    return OK;
  }
  

  int ind;
  int fil = pin->filas;
  int col = pin->columnas;
  pout->tipo = GRISES;
  pout->filas = fil;
  pout->columnas = col;

  uint8_t *ptR = &pin->pixelesR[0];
  uint8_t *ptG = &pin->pixelesG[0];
  uint8_t *ptB = &pin->pixelesB[0];

  for(i = 0; i < fil; i++){
    for(j = 0; j < col; j++){
      ind=i*col+j;
      pout->pixeles[ind] = 0.299*(*ptR++) + 0.587*(*ptG++) + 0.114*(*ptB++);
    }
  }

  DEBPRINT("convertirImagenColorAGris: se retornó sin errores\n");
  return OK;
}
