/**
 * @file imagen.h
 * @author Gregory Randall
 * @date 31 dec 2023
 *
 * Archivo de encabezado de las funciones del obligatorio 1.
 */



#define GAUSSIANO 1
#define SALYPIMIENTA 2
#define PI 3.14159265

#define IMAGEN_H
#include "error.h"

/**
 * @brief Enumerado para indicar si es imagen gris o color.
 */
typedef enum tipo_imagen {GRISES = 0, COLOR = 1} TipoImagen;


/**
 * @brief Tipo de datos para almacenar una imagen.
 *
 * Contiene:
 *
 * TipoImagen tipo: un campo para indicar si se trata de imagen GRIS o COLOR
 * 
 * int columnas,lineas,valorMaximo: para registrar el encabezado de imagen.
 *
 * Pixel* pixeles: puntero a arreglo unidimensional de pixeles.
 * 
 * Ejemplo de uso:
 * @code
 * Imagen testim;
 * testim.pixeles[3*col+4].tipo = COLOR;
 * testim.pixeles[3*col+4].valor.color.rojo = 231;
 * @endcode
 */
typedef struct imagen{
  TipoImagen tipo;
  int columnas;
  int filas;
  uint8_t* pixeles;
  uint8_t* pixelesR;
  uint8_t* pixelesG;
  uint8_t* pixelesB;
} Imagen;


/*------------------FUNCIONES-------------------*/
/**
 * @brief Inicializa una imagen y reserva memoria para sus pixeles.
 *
 * Inicializa los datos de la imagen apuntada por pnew con los parámetros
 * especificados, y reserva memoria para sus nuevos pixeles. El campo
 * valorMaximo si inicializa a 0. Devuelve un valor de tipo CodigoError según
 * el caso. 
 *
 * @param pnew Puntero a imagen a inicializar
 * @param ancho Ancho imagen.
 * @param alto Alto imagen.
 * @param tipo Tipo imagen (color o gris).
 * @return CodigoError.
 */
CodigoError inicializarImagen(Imagen* pnew, int columnas, int filas,
			      TipoImagen tipo);


/**
 * @brief Libera memoria de pixeles y coloca atributos imagen a 0.
 *
 * @param pimg Puntero a imagen a liberar.
 * @return CodigoError.
 */
CodigoError destruirImagen(Imagen* pimg);

/**
 * @brief Copia la inicializacion de la imagen apuntada por pin en pout.
 *
 * Nota: solo se copian los datos de los campos que no son pixeles, para 
 * pixeles únicamente se inicializa la memoria pero no se copia contenido.
 *
 * @param pin Puntero a imagen origen copia.
 * @param pout Puntero a imagen destino copia.
 * @return CodigoError.
 */
CodigoError duplicarImagen(const Imagen* pin, Imagen* pout);

/**
 * @brief Lee el contenido del archivo ubicado en ruta en la imagen apuntada por
 * pimg
 *
 * Es capaz de leer archivos en formato PPM, PPM plano, PGM, PGM plano.
 *
 * @param ruta arreglo de caracteres con la ruta al archivo que se desea leer.
 * @param pin Puntero a imagen para cargar la imagen.
 * @return CodigoError.
 */
CodigoError leerImagen(const char* ruta, Imagen* pimg);

/**
 * @brief Guarda el contenido de la imagen apuntada por pimg en el archivo
 * ubicado en ruta.
 *
 * Es capaz de escribir archivos en formato PPM, PPM plano, PGM, PGM plano.
 *
 * @param pin Puntero a imagen que se desea escribir en archivo.
 * @param ruta Arreglo de caracteres con la ruta al archivo que se desea
 * escribir.
 * @return CodigoError.
 */
CodigoError escribirImagen(const Imagen* pimg, char* ruta);

/**
 * @brief copia los atributos de una imagen de tipo color de entrada apuntada
 * por pin en la imagen de tipo gris apuntada por pout.
 *
 * @param pin Puntero a imagen color
 * @param pout Puntero a imagen blanco y negro, puede no estar inicializado.
 * @return CodigoError.
 */
CodigoError convertirImagenColorAGris(const Imagen* pin, Imagen* pout);

CodigoError AgregarRudioSalYPimienta(const Imagen* pin, Imagen* pout, float potencia, int seed);
CodigoError AgregarRudioGaussiano(const Imagen* pin, Imagen* pout, float potencia, int seed);
double SampleGaussian(double mu, double sigma);
  
CodigoError FiltroMediana(const Imagen* pin, Imagen* pout, int orden);
CodigoError FiltroPromedio(const Imagen* pin, Imagen* pout, int orden);

void Plot_Histo(long int * histo, int nbBins, float factor);

long int * Histo(const Imagen* pin);

/**
 * @brief Genera imagen a traves de una LUT
 *
 * @param pin Puntero a imagen sobre la cual operar
 * @param pout Puntero a imagen de salida
 * @param LUT puntero a Look Up Table
 * @return CodigoError.
 */
CodigoError LUT(const Imagen* pin, Imagen* pout, long int * LUT);

/**
 * @brief Genera el histograma accumulado a partir del histograma
 *
 * @param histoIn Puntero a histograma
 * @param histoAcum puntero a histograma acumulado, allocado fuera de la funcion
 * @param Si vale 1,  sus valores entre 0 y 255, para ser utilizado con LUT.
 * @return CodigoError.
 */
long int * HistoAcum(const long int* histo, int size);
CodigoError imprimirHisto(long int *histo);

CodigoError imprimirRegionFloat(float *img,int fil, int col, int fi,int ff,int ci, int cf);
CodigoError imprimirRegionImg(const Imagen* pin, int fi,int ff,int ci, int cf, int col);
CodigoError imprimirRegionInt(int *img,int fil, int col, int fi,int ff,int ci, int cf);
float *Color2Gris(float* imgRGB,int col, int fil);
